ChatHandler = class(Turbine.UI.Control)
function ChatHandler:Constructor()
    Turbine.UI.Control.Constructor(self)

    self.ChatFunctions = {}
    self.player = Turbine.Gameplay.LocalPlayer.GetInstance()
    self.party = self.player:GetParty()
    self.playerName = self.player:GetName()

    self.callbacks = {}

    self.ChatEvent = function(sender, event, args)
        if type(event)=="function" then
            event(args)
        else
            if type(event)=="table" then
                for i=1,#event do
                    if type(event[i])=="function" then
                        event[i](args)
                    end
                end
            end
        end
    end

    self.UsedChatChannels = {
        [Turbine.ChatType.Raid] = true,
        [Turbine.ChatType.Fellowship] = true,
        [Turbine.ChatType.Say] = true,
        [Turbine.ChatType.Quest] = true,
        [Turbine.ChatType.Standard] = true,
        [Turbine.ChatType.Death] = true,
        [Turbine.ChatType.EnemyCombat] = true,
        [Turbine.ChatType.Advancement] = true,
        [Turbine.ChatType.PlayerCombat] = true,
    }

    self.cleanPartyMessage = function(message)

        local name, cleanedLine
        -- Fellowship
        if string.find(message, "%[To Fellowship%]") then
            cleanedLine = string.match(message, "%[To Fellowship%] (.*)")
        elseif string.find(message, "%[Fellowship%]") then
            name, cleanedLine = string.match(message, "%[Fellowship%] (.*): (.*)")
        -- Group
        elseif string.find(message, "%[To Group%]") then
            cleanedLine = string.match(message, "%[To Group%] (.*)")
        elseif string.find(message, "%[Group%]") then
            name, cleanedLine = string.match(message, "%[Group%] (.*): (.*)")
        -- Raid
        elseif string.find(message, "%[To Raid%]") then
            cleanedLine = string.match(message, "%[To Raid%] (.*)")
        elseif string.find(message, "%[Raid%]") then
            name, cleanedLine = string.match(message, "%[Raid%] (.*): (.*)")
        end

        if name then 
            return string.match(name, ">(.*)<\\Select>"), cleanedLine
        else
            return self.player:GetName(), cleanedLine
        end
    end

    -- Chat handling functions
    self.PartyChatHandling = function(message)
        -- Handle impossible messages :)
        if self.party == nil then return end

        -- Check name of chat sender
        --local senderName = self:ParseName(message)
        local senderName, cleanedLine = self.cleanPartyMessage(message)
        if cleanedLine then cleanedLine = cleanedLine:gsub("\n", "") end

        self.ChatEvent(self, self.Party, cleanedLine)
    end

    self.ChatFunctions[Turbine.ChatType.Fellowship] = self.PartyChatHandling
    self.ChatFunctions[Turbine.ChatType.Raid] = self.PartyChatHandling

    self.SayChatHandling = function(message)
        self.ChatEvent(self, self.Say, message)
        self.ChatEvent(self, self.SayLine, self:ParseTextLineOnly(message))
    end

    self.ChatFunctions[Turbine.ChatType.Say] = self.SayChatHandling

    self.QuestChatHandling = function(message)
        -- Instance controller handling
        --if InstanceController then InstanceController:CheckInstanceState(message) end

        self.ChatEvent(self, self.QuestText, {["Line"] = message})

        if string.find(message, "New Quest") then
            local questName = string.match(message, "New Quest: (.*)")
            if questName then
                self.ChatEvent(self, self.QuestAccepted, {["Name"] = questName})
                return
            end
        end

        if string.find(message, "New Deed") then
            local deedName = string.match(message, "New Deed: (.*)")
            if deedName then
                self.ChatEvent(self, self.NewDeed, {["Name"] = deedName})
                return
            end
        end

        if string.find(message, "Completed:") then
            local questName = string.match(message, "Completed:\n(.*)\n")
            if questName then
                self.ChatEvent(self, self.QuestOrDeedCompleted, {["Name"] = questName})
                return
            end
        end

        if string.match(message, "(.*) %((%d+)/(%d+)%)") then
            local name, progress, total = string.match(message, "(.*) %((%d+)/(%d+)%)")
            self.ChatEvent(self, self.QuestOrDeedProgress, {["Name"] = name, ["Progress"] = progress, ["Total"] = total})
            return
        end
    end
    self.ChatFunctions[Turbine.ChatType.Quest] = self.QuestChatHandling

    -- Lootlock

    self.StandardChatHandling = function(message)
        self.ChatEvent(self, self.Standard, message)

        if string.find(message, "<rgb=#FF9900><Prime></rgb> ") then return end

        -- Instancecontroller chat handling
        --if InstanceController then InstanceController:CheckInstanceState(message) end

        -- Handle party change reloading
        if string.find(message, "[Ff]ellowship") or string.find(message, "[Rr]aid") or string.find(message, "[Dd]ismiss")  then
            if string.find( message, " has joined your Fellowship." ) or 
            string.find( message, " has joined your Raid." ) or
            string.find( message, "You have joined a Fellowship." ) or
            string.find( message, "You have joined a Raid." ) or
            string.find( message, " has left your Fellowship." ) or
            string.find( message, " has left your Raid." ) or
            string.find( message, "You leave your Fellowship." ) or
            string.find( message, "You leave your Raid." ) or
            string.find( message, "You dismiss " ) or
            string.find( message, " from the Raid." ) or
            string.find( message, " has been dismissed from your Fellowship." ) or
            string.find( message, " has been dismissed from your Raid." ) or
            string.find( message, "You have been dismissed from your Raid." ) or
            string.find( message, "You have been dismissed from your Fellowship." ) or
            string.find( message, "Your Fellowship has been disbanded." ) or
            string.find( message, "Your Raid has been disbanded." ) or
            string.find( message, "You have joined the chat room for your raid, raid chat is now available." ) or
            string.find( message, "You have joined the chat room for your Fellowship, Fellowship chat is now available." ) or
            string.find( message, "the leader of the raid") or
            string.find( message, "the leader of the fellowship") then
                self.ChatEvent(self, self.PartyChange, {})
            end
        end
    end

    self.ChatFunctions[Turbine.ChatType.Standard] = self.StandardChatHandling

    self.DeathChatHandling = function(message)

        -- Player death
        if string.find(message, "You succumb to your wounds.") then 
            self.ChatEvent(self, self.Retreat, {["Name"] = self.player:GetName()})
        elseif string.find(message, "Your fellow,") then
            local retreatName = string.match(message, "Your fellow, (.*) has been defeated!")
            if retreatName then
                self.ChatEvent(self, self.Retreat, {["Name"] = retreatName})
            end
        end

        -- KB handling
        if string.find(message, "Your mighty blow defeated") then
            local kbName = string.gsub(string.match(message, "Your mighty blow defeated (.+)%."),"^[Tt]he ","")
            self.ChatEvent(self, self.Defeated, {["Name"] = kbName})
        elseif string.find(message, "defeated") then
            local source, name = string.match(message, "(.*) defeated (.+)%.")
            if name then name = string.gsub(name, "^[Tt]he ","") end
            self.ChatEvent(self, self.Defeated, {["Source"] = source, ["Name"] =  name or string.match(message,"(.*) has been defeated.")})
        elseif string.find(message, "Your mighty blow topples") then
            --StatEventHandling:AddKB(string.match(message, "Your mighty blow topples (.+)%."), true)
            -- KB handling
        end

        -- Getting killed handling
        if string.find(message, " incapacitated you.") then
            local source = string.match(message, "(.+) incapacitated you.")
            self.ChatEvent(self, self.Defeated, {["Death_Source"] = source, ["Self"] = true, ["Name"] =  self.player:GetName()})
        end

        -- Misadventure death handling
        if string.find(message, "You have been incapacitated by misadventure.") then
            self.ChatEvent(self, self.Defeated, {["Death_Source"] = "Misadventure", ["Self"] = true, ["Name"] =  self.player:GetName()})
        end
    end

    self.ChatFunctions[Turbine.ChatType.Death] = self.DeathChatHandling

    self.EnemyCombatChatHandling = function(message)
        --local source, _, skillName, rest = string.match(string.gsub(string.gsub(message,"<rgb=#......>(.*)</rgb>","%1"),"^%s*(.-)%s*$", "%1"),"^(.*) scored a (.*)hit with (.*) on (.*) for (.-) (.*)%.$")
        local source, _, skillName, rest = string.match(string.gsub(string.gsub(message,"<rgb=#......>(.*)</rgb>","%1"),"^%s*(.-)%s*$", "%1"),"^(.*) scored a (.*)hit with (.*) on (.*)%.$")
        if skillName then
            if rest then
                local target, _ = string.match(rest,"^(.*) for (.*)$")
                if target then
                    self.ChatEvent(self, self.EnemyCombat, {["SkillName"] = skillName, ["Target"] = string.gsub(target, "^[Tt]he ", "")})
                else
                    self.ChatEvent(self, self.EnemyCombat, {["SkillName"] = skillName, ["Target"] = string.gsub(rest, "^[Tt]he ", "")})
                end
                
            else
                self.ChatEvent(self, self.EnemyCombat, {["SkillName"] = skillName, ["Target"] = ""})
            end
        end  
    end

    self.ChatFunctions[Turbine.ChatType.EnemyCombat] = self.EnemyCombatChatHandling

    self.PlayerCombatChatHandler = function(message)
        local cleanedMessage = string.gsub(string.gsub(message,"<rgb=#......>(.*)</rgb>","%1"),"^%s*(.-)%s*$", "%1")

        if string.find(cleanedMessage, "hit") then
            --local source,_,skillName,target,_ = string.match(cleanedMessage,"^(.*) scored a (.*)hit with (.*) on (.*) for (.-) (.*)%.$")
            local source, _, skillName, rest = string.match(cleanedMessage,"^(.*) scored a (.*)hit with (.*) on (.*)%.$")

            if skillName then
                if rest then
                    local target, _ = string.match(rest,"^(.*) for (.*)$")
                    if target then
                        self.ChatEvent(self, self.PlayerCombat, {["Source"] = source, ["SkillName"] = skillName, ["Target"] = target})
                    else
                        self.ChatEvent(self, self.PlayerCombat, {["Source"] = source, ["SkillName"] = skillName, ["Target"] = string.gsub(rest, "^[Tt]he ", "")})
                    end
                    
                else
                    self.ChatEvent(self, self.PlayerCombat, {["Source"] = source, ["SkillName"] = skillName, ["Target"] = ""})
                end
                return
            end  
        elseif string.find(cleanedMessage, "heal") then
            local skillName, _, target, _ = string.match(cleanedMessage,"^(.*) applied a (.*)heal to (.*) restoring (.*) points to Morale%.$")
            if skillName then
                self.ChatEvent(self, self.PlayerHeal, {["SkillName"] = skillName, ["Target"] = target})
                return
            end 
            
            -- source, type, skillName, target, number
            local source, _, skillName, target, _ = string.match(cleanedMessage, "^(.*) applied a (.*)heal with (.*) to (.*) restoring (.*) points to Morale%.$")
            if skillName then
                self.ChatEvent(self, self.PlayerHeal, {["Source"] = source, ["SkillName"] = skillName, ["Target"] = target})
                return
            end
        elseif string.find(cleanedMessage, "benefit") then
            local source, _, skillName, target = string.match(cleanedMessage,"^(.*) applied a (.*)benefit with (.*) on (.*)%.$")
            if skillName then
                self.ChatEvent(self, self.PlayerBenefit, {["SkillName"] = skillName, ["Target"] = target, ["Source"] = source})
                return
            end 
        elseif string.find(cleanedMessage, "dispelled") then
            local effectName, target = string.match(cleanedMessage, "You have dispelled (.*) from (.*).")
            if effectName then
                self.ChatEvent(self, self.PlayerDispell, {["EffectName"] = effectName, ["Target"] = target})
                return
            end
        end
    end

    self.ChatFunctions[Turbine.ChatType.PlayerCombat] = self.PlayerCombatChatHandler

    self.AdvancementHandling = function(message)
        if string.find(message, "You have acquired") then
            self.ChatEvent(self, self.Advancement, {["Message"] = message})
        end
    end

    self.ChatFunctions[Turbine.ChatType.Advancement] = self.AdvancementHandling

    -- Set parse function
    self.callbacks["Received"] = AddCallback( Turbine.Chat, "Received", function(sender, args)
        if self.ChatFunctions[args.ChatType] then
            self.ChatFunctions[args.ChatType](args.Message)
        end
    end)
end

function ChatHandler:RemoveCallbacks()
    for key, value in pairs(self.callbacks) do
        RemoveCallback(Turbine.Chat, key, value)
    end
end

function ChatHandler:ParseTextLineOnly(message)
    local _, _, line = string.match(message, ">(.*)<\\Select> (.*), ''(.*)''")
    if line then
        return line
    end

    local _, _, line = string.match(message, ">(.*)<\\Select> (.*), '(.*)'")
    if line then
        return line
    end

    if string.find(message, "You say, ") then
        return string.match(message, "You say, '(.*)'") or ""
    end

    return message
end

function ChatHandler:ParseName(message)
    local name = string.match(message, ">(.*)<\\Select>")
    return name or ""
end

function ChatHandler:ParsePartyName(message)
    local name = string.match(message, ">(.*)<\\Select>")
    return name or ""
end